﻿using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;


//====================================================
// Описание работы классов и методов исходника на:
// https://www.interestprograms.ru
// Исходные коды программ и игр
//====================================================

namespace WpfMeasureText.MyCode
{
    internal class DrawCanvas : Canvas
    {

        public delegate void ChangeDrawTextEventHandler(ChangeDrawTextEventArgs args);

        public event ChangeDrawTextEventHandler ChangeDraw;

        public virtual void OnChangeDraw(ChangeDrawTextEventArgs args)
        {
            ChangeDraw?.Invoke(args);
        }

        private string _text = "Measuring \n text";
        private double _fontSize = 12;
        private double _lineHeight;
        private string _familyName = "Arial";
        private FontStyle _fontStyle = FontStyles.Normal;


        public string Text
        {
            get => _text;
            set
            {
                _text = value;
                InvalidateVisual();
            }
        }

        public double FontSize
        {
            get => _fontSize;
            set
            {
                _fontSize = value;
                InvalidateVisual();
            }
        }


        public double LineHeight
        {
            get => _lineHeight;
            set
            {
                _lineHeight = value;
                InvalidateVisual();
            }
        }


        public string FamilyName
        {
            get => _familyName;
            set
            {
                _familyName = value;
                InvalidateVisual();
            }
        }



        public FontStyle FontStyle
        {
            get => _fontStyle;
            set
            {
                _fontStyle = value;
                InvalidateVisual();
            }
        }




        protected override void OnRender(DrawingContext dc)
        {

            FontFamily ff = new(_familyName);
            FontWeight fw = FontWeights.Bold;
            FontStretch fsh = FontStretches.Normal;

            double pixelsPerDip = VisualTreeHelper.GetDpi(this).PixelsPerDip;
            Typeface typeface = new(ff, _fontStyle, fw, fsh);

            FormattedText formattedText = new(
                _text,
                System.Globalization.CultureInfo.CurrentCulture,
                FlowDirection.LeftToRight,
                typeface,
                _fontSize,
                Brushes.Black,
                pixelsPerDip
                );

            formattedText.LineHeight = _lineHeight;
            formattedText.TextAlignment = TextAlignment.Left;

            double scale = 1;
            ScaleTransform st = new(scale, scale);
            dc.PushTransform(st);

            formattedText.MaxTextWidth = this.ActualWidth / scale;
            formattedText.MaxTextHeight = this.ActualHeight / scale;






            // Текст расположим в середине канвы.
            double x = this.ActualWidth / 2 - (formattedText.OverhangLeading + formattedText.Width - formattedText.OverhangTrailing) / 2;
            double y = 0;
            dc.DrawText(formattedText, new Point(x, y));


            // Высота всех строк 
            double heightText = y + formattedText.Height;


            // Экземпляр для измерений количества строк, высоты строки и др.
            FormattedText addFormattedText = new(
                "A",
                System.Globalization.CultureInfo.CurrentCulture,
                FlowDirection.LeftToRight,
                typeface,
                _fontSize,
                Brushes.Black,
                pixelsPerDip
                );

            double lineheight = formattedText.LineHeight == 0 ? addFormattedText.Height : formattedText.LineHeight;

            int numlines = (int)formattedText.Height / (int)lineheight;


            // Передача данных для отображения их в текстовых метках.
            ChangeDrawTextEventArgs args = new()
            {
                OverhangLeading = formattedText.OverhangLeading,
                Width = formattedText.Width,
                OverhangTrailing = formattedText.OverhangTrailing,

                Baseline = formattedText.Baseline,
                Height = formattedText.Height,
                Extent = formattedText.Extent,
                OverhangAfter = formattedText.OverhangAfter,
                LineHeight = lineheight,
                NumberLines = numlines
            };

            OnChangeDraw(args);



            //---- horizontal ---

            // Корректирующая для ведущего и хвостового выступа: чтобы визуально не сливались с прямоугольником ширины-высоты.
            double correctOverhangHeight = formattedText.Height / 10;

            // Визуальное отображение ведущего выступа
            // От начала рисования до начала строки по горизонтали (это нахождение начальной точки выравнивания текста по горизонтали)
            dc.DrawLine(new Pen(new SolidColorBrush(Color.FromArgb(150, 0, 180, 0)), heightText - correctOverhangHeight),
                                 new Point(x, y + (heightText / 2)), new Point(x + formattedText.OverhangLeading, y + (heightText / 2)));



            // Визуальное отображение ширины и высоты строк.
            Rect rectWidthText = new(
                new Point(x, y),
                new Point(x + formattedText.Width, y + formattedText.Height)
                );

            dc.DrawRectangle(new SolidColorBrush(Color.FromArgb(50, 255, 0, 0)), null, rectWidthText);



            // Визуальное отображение хвостового выступа
            dc.DrawLine(new Pen(new SolidColorBrush(Color.FromArgb(100, 0, 0, 200)), heightText - correctOverhangHeight),
                new Point(x + formattedText.Width, y + (heightText / 2)), new Point(x + formattedText.Width - formattedText.OverhangTrailing, y + (heightText / 2)));


            // Линия актуальной ширины пикселей текста.
            dc.DrawLine(new Pen(new SolidColorBrush(Color.FromArgb(200, 120, 170, 170)), 10),
                new Point(x + formattedText.OverhangLeading, y - 5), new Point(x + formattedText.Width - formattedText.OverhangTrailing, y - 5));

            //---- horizontal ---


            // --- vertical ---


            // Базовая линия текста, по нижним частям основных букв, без свисающих хвостов.
            // По нижней части большинства прописных букв.
            dc.DrawLine(new Pen(Brushes.Yellow, 2), new Point(x, y + formattedText.Baseline + 1), new Point(x + formattedText.Width, y + formattedText.Baseline + 1));



            // Extent - пространство занимемое пикселями текста.
            // Это фактическая высота символов во всех строках текста, актуальное пространство занимаемое только буквами, без пустых пространств.
            // Расстояние от самого верхнего рисуемого пикселя первой строки до самого нижнего рисуемого пикселя последней строки.
            // Так как метки верхнего рисуемого пикселя у нас нет, но есть метка нижнего рисуемого пикселя.

            // Находим нижний пиксель
            // низ последней строки - fmtText.Height
            // расстояние от низа строки до нижнего пискеля - fmtText.OverhangAfter
            // Поскольку значение fmtText.OverhangAfter является положительным, если самый нижний рисуемый пиксел находится под нижней границей строки и
            // является отрицательным, если он находится в пределах строки.
            // Чтобы найти границу самого нижнего рисуемого пикселя, необходимо к высоте текста прибавить fmtText.OverhangAfter:
            double maxbottompixel = y + formattedText.Height + formattedText.OverhangAfter;

            // затем от самого нижнего пикселя получаем расстояние до самого верхнего пикселя при помощи значения fmtText.Extent:
            double maxtoppixel = maxbottompixel - formattedText.Extent;

            // В итоге получаем высоту только рисованных пикселей всех строк текста.
            Rect rectExtent = new(x + formattedText.Width / 2 - 25, maxtoppixel, 50, maxbottompixel - maxtoppixel);
            dc.DrawRectangle(new SolidColorBrush(Color.FromArgb(150, 150, 150, 0)), null, rectExtent);



            // Наглядное изображение высоты каждой строки.
            for (int i = 0; i < numlines; i++)
            {
                Rect rectLineHeight = new(x + formattedText.Width / 2 - (20 * (i + 0)), y + (i * lineheight), 20, y + lineheight);
                dc.DrawRectangle(new SolidColorBrush(Color.FromArgb(180, 0, 0, 255)), null, rectLineHeight);
            }


            // Расстояние от низа последней строки до самых нижний пикселей текста.
            dc.DrawLine(new Pen(new SolidColorBrush(Color.FromArgb(150, 100, 100, 100)), 200),
                new Point(x + formattedText.Width / 2, y + formattedText.Height), new Point(x + formattedText.Width / 2, y + formattedText.Height + formattedText.OverhangAfter));


            // --- vertical ---


            // Поле канвы.
            Rect r = new(0, 0, this.ActualWidth, this.ActualHeight);
            dc.DrawRectangle(null, new Pen(new SolidColorBrush(Colors.Chocolate), 1), r);

        }
    }
}
